#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/ioctl.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>

#include "buffer.h"
#include "server.h"
#include "keyvalue.h"
#include "log.h"

#include "http_chunk.h"
#include "fdevent.h"
#include "connections.h"
#include "response.h"
#include "joblist.h"

#include "plugin.h"

#include <fastcgi.h>
#include <stdio.h>

#ifdef HAVE_SYS_FILIO_H
# include <sys/filio.h>
#endif


/*
 * 
 * TODO:
 * 
 * - add timeout for a connect to a non-fastcgi process
 *   (use state_timestamp + state)
 * 
 */
typedef struct {
	size_t *ptr;
	size_t used;
	size_t size;
} buffer_int;

typedef struct {
	buffer_int fcgi_request_id;
	
	buffer *fcgi_env;
	buffer *fcgi_master;
	
	array *extensions;
	
	buffer *path;
	buffer *parse_response;
	
	int debug;
} plugin_data;

INIT_FUNC(mod_fastcgi_init) {
	plugin_data *p;
	
	p = malloc(sizeof(plugin_data));
	memset(p, 0, sizeof(plugin_data));
	
	p->fcgi_env = buffer_init();
	p->fcgi_master = buffer_init();
	
	p->path = buffer_init();
	p->parse_response = buffer_init();
	
	p->extensions = array_init();
	
	return p;
}


FREE_FUNC(mod_fastcgi_free) {
	plugin_data *p = p_d;
	buffer_int *r = &(p->fcgi_request_id);
	
	UNUSED(srv);

	if (r->ptr) free(r->ptr);
	
	buffer_free(p->fcgi_env);
	buffer_free(p->fcgi_master);
	buffer_free(p->path);
	buffer_free(p->parse_response);
	
	array_free(p->extensions);
	
	free(p);
	
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC(mod_fastcgi_set_defaults) {
	plugin_data *p = p_d;
	data_unset *du;
	size_t i = 0;
	
	config_values_t cv[] = { 
		{ "fastcgi.server",              NULL, T_CONFIG_LOCAL },       /* 0 */
		{ "fastcgi.debug",               NULL, T_CONFIG_SHORT },       /* 1 */
		{ NULL,                          NULL, T_CONFIG_UNSET }
	};
	
	cv[i++].destination = p->extensions;
	cv[i++].destination = &(p->debug);
	
	if (0 != config_insert_values(srv, cv)) {
		return HANDLER_ERROR;
	}
	
	if (NULL != (du = array_get_element(srv->config, "fastcgi.server"))) {
		if (du->type == TYPE_ARRAY) {
			size_t j;
			data_array *da = (data_array *)du;
			
			/* get the extension from -> key */
			
			for (j = 0; j < da->value->used; j++) {
				if (da->value->data[j]->type == TYPE_ARRAY) {
					size_t n;
					data_array *da_ext = (data_array *)da->value->data[j];
			
					/* 
					 * da_ext->key == name of the extension 
					 */
					
					for (n = 0; n < da_ext->value->used; n++) {
						if (da_ext->value->data[n]->type == TYPE_ARRAY) {
							size_t m;
							data_array *da_host = (data_array *)da_ext->value->data[n];
							const char *host = NULL, *docroot = NULL;
							unsigned short port = 0;
							
							for (m = 0; m < da_host->value->used; m++) {
								data_unset *da_unset_entry = da_host->value->data[m];
								
								switch(da_unset_entry->type) {
								case TYPE_STRING: {
									data_string *da_entry = (data_string *)da_unset_entry;
									
									if (0 == strcmp(da_entry->key->ptr, "host")) {
										host = da_entry->value->ptr;
									} else if (0 == strcmp(da_entry->key->ptr, "docroot")) {
										docroot = da_entry->value->ptr;
									} else {
										log_error_write(srv, __FILE__, __LINE__, "sbbbb", 
												"(warning) unexpected key (string):", 
												da->key,
												da_ext->key,
												da_host->key,
												da_entry->key);
									}
									break;
								}
								case TYPE_INTEGER: {
									data_integer *da_entry = (data_integer *)da_unset_entry;
									if (0 == strcmp(da_entry->key->ptr, "port")) {
										port = da_entry->value;
									} else {
										log_error_write(srv, __FILE__, __LINE__, "sbbbb", 
												"(warning) unexpected key (short):", 
												da->key,
												da_ext->key,
												da_host->key,
												da_entry->key);
									}
									break;
								}
								default:
									log_error_write(srv, __FILE__, __LINE__, "sbbbb", 
												"(warning) unexpected key (short):", 
												da->key,
												da_ext->key,
												da_host->key,
												da_unset_entry->key);
								}
							}
							
							if (host == NULL) {
								log_error_write(srv, __FILE__, __LINE__, "sbbbs", 
										"missing key (string):", 
										da->key,
										da_ext->key,
										da_host->key,
										"host");
								
								return HANDLER_ERROR;
							} else if (port == 0) {
								log_error_write(srv, __FILE__, __LINE__, "sbbbs", 
										"missing key (short):", 
										da->key,
										da_ext->key,
										da_host->key,
										"port");
								return HANDLER_ERROR;
							} else {
								data_fastcgi *df = data_fastcgi_init();
								data_array *dfa; 
								
								buffer_copy_string_buffer(df->key, da_host->key);
								buffer_copy_string(df->host, host);
								df->port = port;
								
								if (docroot) buffer_copy_string(df->docroot, docroot);
								
								/* if extension already exists, take it */
								
								if (NULL == (dfa = (data_array *)array_get_element(p->extensions, da_ext->key->ptr))) {
									dfa = data_array_init();
									
									buffer_copy_string_buffer(dfa->key, da_ext->key);
									
									array_insert_unique(dfa->value, (data_unset *)df);
									array_insert_unique(p->extensions, (data_unset *)dfa);
								} else {
									array_insert_unique(dfa->value, (data_unset *)df);
								}
							}
							
						} else {
							log_error_write(srv, __FILE__, __LINE__, "ssSBS", 
									"unexpected type for key:", 
									"fastcgi.server", 
									"[", da_ext->value->data[n]->key, "](string)");
							
							return HANDLER_ERROR;
						}
					}
				} else {
					log_error_write(srv, __FILE__, __LINE__, "sssbs", 
							"unexpected type for key: ", "fastcgi.server", 
							"[", da->value->data[j]->key, "](string)");
					
					return HANDLER_ERROR;
				}
			}
		} else {
			log_error_write(srv, __FILE__, __LINE__, "sss", "unexpected type for key: ", "fastcgi.server", "array of strings");
			
			return HANDLER_ERROR;
		}
	}
	
	return HANDLER_GO_ON;
}


static int fcgi_requestid_new(server *srv, plugin_data *p) {
	size_t m = 0;
	size_t i;
	buffer_int *r = &(p->fcgi_request_id);
	
	UNUSED(srv);

	for (i = 0; i < r->used; i++) {
		if (r->ptr[i] > m) m = r->ptr[i];
	}
	
	if (r->size == 0) {
		r->size = 16;
		r->ptr = malloc(sizeof(int) * r->size);
	} else if (r->used == r->size) {
		r->size += 16;
		r->ptr = realloc(r->ptr, sizeof(int) * r->size);
	}
	
	r->ptr[r->used++] = ++m;
	
	return m;
}

static int fcgi_requestid_del(server *srv, plugin_data *p, size_t request_id) {
	size_t i;
	buffer_int *r = &(p->fcgi_request_id);
	
	UNUSED(srv);

	for (i = 0; i < r->used; i++) {
		if (r->ptr[i] == request_id) break;
	}
	
	if (i != r->used) {
		/* found */
		
		if (i != r->used - 1) {
			r->ptr[i] = r->ptr[r->used - 1];
		}
		r->used--;
	}
	
	return 0;
}


static int fcgi_env_add(buffer *env, const char *key, size_t key_len, const char *val) {
	size_t val_len, len;
	
	if (!key || !val) return -1;
	
	val_len = strlen(val);
	
	len = key_len + val_len;
	
	len += key_len > 127 ? 1 : 4;
	len += val_len > 127 ? 1 : 4;
	
	buffer_prepare_append(env, len);
	
	if (key_len > 127) {
		env->ptr[env->used++] = ((key_len >> 24) & 0xff) | 0x80;
		env->ptr[env->used++] = (key_len >> 16) & 0xff;
		env->ptr[env->used++] = (key_len >> 8) & 0xff;
		env->ptr[env->used++] = (key_len >> 0) & 0xff;
	} else {
		env->ptr[env->used++] = (key_len >> 0) & 0xff;
	}
	
	if (val_len > 127) {
		env->ptr[env->used++] = ((val_len >> 24) & 0xff) | 0x80;
		env->ptr[env->used++] = (val_len >> 16) & 0xff;
		env->ptr[env->used++] = (val_len >> 8) & 0xff;
		env->ptr[env->used++] = (val_len >> 0) & 0xff;
	} else {
		env->ptr[env->used++] = (val_len >> 0) & 0xff;
	}
	
	memcpy(env->ptr + env->used, key, key_len);
	env->used += key_len;
	memcpy(env->ptr + env->used, val, val_len);
	env->used += val_len;
	
	return 0;
}

static int fcgi_header(FCGI_Header * header, unsigned char type, size_t request_id, int contentLength, unsigned char paddingLength) {
	header->version = FCGI_VERSION_1;
	header->type = type;
	header->requestIdB0 = request_id & 0xff;
	header->requestIdB1 = (request_id >> 8) & 0xff;
	header->contentLengthB0 = contentLength & 0xff;
	header->contentLengthB1 = (contentLength >> 8) & 0xff;
	header->paddingLength = paddingLength;
	header->reserved = 0;
	
	return 0;
}

static int fcgi_establish_connection(server *srv, connection *con, void *p_d, data_fastcgi *host, int fcgi_fd) {
	struct sockaddr_in fcgi_addr;
	socklen_t servlen;
	plugin_data *p = p_d;
	
	memset(&fcgi_addr, 0, sizeof(fcgi_addr));

	fcgi_addr.sin_family = AF_INET;
	fcgi_addr.sin_addr.s_addr = inet_addr(host->host->ptr);
	fcgi_addr.sin_port = htons(host->port);
	servlen = sizeof(fcgi_addr);
	
	if (-1 == connect(fcgi_fd, (struct sockaddr *) &fcgi_addr, servlen)) {
		if (errno == EINPROGRESS || errno == EALREADY) {
			if (p->debug) {
				log_error_write(srv, __FILE__, __LINE__, "sd", 
						"connect delayed:", fcgi_fd);
			}
			
			return -fcgi_fd;
		} else {
			int oerrno = errno;
			con->fcgi.fde_ndx = fdevent_event_del(srv->ev, con->fcgi.fde_ndx, con->fcgi.fd);
			fdevent_fdc_del(srv->ev->fdc, fcgi_fd);
			close(fcgi_fd);

			errno = oerrno;
			log_error_write(srv, __FILE__, __LINE__, "sdsd", 
					"connect failed:", fcgi_fd, strerror(errno), errno);
			
			return -1;
		}
	}
	if (p->debug) {
		log_error_write(srv, __FILE__, __LINE__, "sd", 
				"connect succeeded: ", fcgi_fd);
	}
	return fcgi_fd;
}

static int fcgi_env_add_request_headers(server *srv, connection *con, void *p_d) {
	plugin_data *p = p_d;
	size_t i;
	
	for (i = 0; i < con->request.headers->used; i++) {
		data_string *ds;
		
		ds = (data_string *)con->request.headers->data[i];
		
		if (ds->value->used && ds->key->used) {
			size_t j;
			buffer_reset(srv->tmp_buf);
			
			/* don't forward the Authorization: Header */
			if (0 == strcasecmp(ds->key->ptr, "AUTHORIZATION")) {
				continue;
			}
			
			if (0 != strcasecmp(ds->key->ptr, "CONTENT-TYPE")) {
				buffer_copy_string(srv->tmp_buf, "HTTP_");
				srv->tmp_buf->used--;
			}
			
			buffer_prepare_append(srv->tmp_buf, ds->key->used + 2);
			for (j = 0; j < ds->key->used - 1; j++) {
				srv->tmp_buf->ptr[srv->tmp_buf->used++] = 
					isalpha((unsigned char)ds->key->ptr[j]) ? 
					toupper((unsigned char)ds->key->ptr[j]) : '_';
			}
			srv->tmp_buf->ptr[srv->tmp_buf->used] = '\0';
			
			fcgi_env_add(p->fcgi_env, srv->tmp_buf->ptr, srv->tmp_buf->used, ds->value->ptr);
		}
	}
	
	return 0;
}


static int fcgi_create_env(server *srv, connection *con, void *p_d, data_fastcgi *host, size_t request_id) {
	plugin_data *p = p_d;
	FCGI_BeginRequestRecord beginRecord;
	FCGI_Header header;
	
	char buf[32];
	size_t offset;
	
#ifdef HAVE_IPV6
	int ndx, i;
#endif

#define CONST_STRING(x) \
		x, sizeof(x)-1
	
	/* send FCGI_BEGIN_REQUEST */
	
	fcgi_header(&(beginRecord.header), FCGI_BEGIN_REQUEST, request_id, sizeof(beginRecord.body), 0);
	beginRecord.body.roleB0 = FCGI_RESPONDER;
	beginRecord.body.roleB1 = 0;
	beginRecord.body.flags = 0;
	memset(beginRecord.body.reserved, 0, sizeof(beginRecord.body.reserved));
	
	buffer_copy_memory(con->fcgi.write_buffer, (const char *)&beginRecord, sizeof(beginRecord));
	
	/* send FCGI_PARAMS */
	buffer_prepare_copy(p->fcgi_env, 1024);

	if (p->fcgi_master->used == 0) {
#ifdef HAVE_IPV6
		char b2[INET6_ADDRSTRLEN + 1];
#endif
#ifdef PACKAGE_NAME
		fcgi_env_add(p->fcgi_env, CONST_STRING("SERVER_SOFTWARE"), PACKAGE_NAME"/"PACKAGE_VERSION);
#else
		fcgi_env_add(p->fcgi_env, CONST_STRING("SERVER_SOFTWARE"), PACKAGE"/"VERSION);
#endif
		fcgi_env_add(p->fcgi_env, CONST_STRING("SERVER_NAME"),
#ifdef HAVE_IPV6
			     inet_ntop(srv->srv_addr.plain.sa_family, 
				       srv->srv_addr.plain.sa_family == AF_INET6 ? 
				       (const void *) &(srv->srv_addr.ipv6.sin6_addr) :
				       (const void *) &(srv->srv_addr.ipv4.sin_addr),
				       b2, sizeof(b2)-1)
#else
			     inet_ntoa(srv->srv_addr.ipv4.sin_addr)
#endif
			     );
		fcgi_env_add(p->fcgi_env, CONST_STRING("GATEWAY_INTERFACE"), "CGI/1.1");
		
		buffer_prepare_copy(p->fcgi_master, p->fcgi_env->used);
		
		memcpy(p->fcgi_master->ptr, p->fcgi_env->ptr, p->fcgi_env->used);
		p->fcgi_master->used = p->fcgi_env->used;
	} else {
		memcpy(p->fcgi_env->ptr, p->fcgi_master->ptr, p->fcgi_master->used);
		p->fcgi_env->used = p->fcgi_master->used;
	}
	
	ltostr(buf, 
#ifdef HAVE_IPV6
	       ntohs(srv->srv_addr.plain.sa_family ? srv->srv_addr.ipv6.sin6_port : srv->srv_addr.ipv4.sin_port)
#else
	       ntohs(srv->srv_addr.ipv4.sin_port)
#endif
	       );
	
	fcgi_env_add(p->fcgi_env, CONST_STRING("SERVER_PORT"), buf);
	
#ifdef HAVE_IPV6
	ndx = 0;
	for (i = 0; i < INET_NTOP_CACHE_MAX; i++) {
		if (srv->inet_ntop_cache[i].ts != 0) {
			if (srv->inet_ntop_cache[i].family == AF_INET6 &&
			    0 == memcmp(srv->inet_ntop_cache[i].addr.ipv6.s6_addr, con->dst_addr.ipv6.sin6_addr.s6_addr, 16)) {
				/* IPv6 found in cache */
				break;
			} else if (srv->inet_ntop_cache[i].family == AF_INET &&
				   srv->inet_ntop_cache[i].addr.ipv4.s_addr == con->dst_addr.ipv4.sin_addr.s_addr) {
				/* IPv4 found in cache */
				break;
				
			}
		}
	}

	if (i == INET_NTOP_CACHE_MAX) {
		/* not found in cache */
		i = ndx;
		inet_ntop(con->dst_addr.plain.sa_family, 
			  con->dst_addr.plain.sa_family == AF_INET6 ? 
			  (const void *) &(con->dst_addr.ipv6.sin6_addr) :
			  (const void *) &(con->dst_addr.ipv4.sin_addr),
			  srv->inet_ntop_cache[i].b2, INET6_ADDRSTRLEN);
		
		srv->inet_ntop_cache[i].ts = srv->cur_ts;
		srv->inet_ntop_cache[i].family = con->dst_addr.plain.sa_family;
		
		if (srv->inet_ntop_cache[i].family == AF_INET) {
			srv->inet_ntop_cache[i].addr.ipv4.s_addr = con->dst_addr.ipv4.sin_addr.s_addr;
		} else if (srv->inet_ntop_cache[i].family == AF_INET6) {
			memcpy(srv->inet_ntop_cache[i].addr.ipv6.s6_addr, con->dst_addr.ipv6.sin6_addr.s6_addr, 16);
		}
	}
	fcgi_env_add(p->fcgi_env, CONST_STRING("REMOTE_ADDR"),
		     srv->inet_ntop_cache[i].b2);
#else
	fcgi_env_add(p->fcgi_env, CONST_STRING("REMOTE_ADDR"),
		     inet_ntoa(con->dst_addr.ipv4.sin_addr));
#endif
	
	if (con->authed_user->used) {
		fcgi_env_add(p->fcgi_env, CONST_STRING("REMOTE_USER"),
			     con->authed_user->ptr);
	}
	fcgi_env_add(p->fcgi_env, CONST_STRING("REMOTE_ADDR"),
		     srv->inet_ntop_cache[i].b2);
	
	/* request.content_length < SSIZE_MAX, see request.c */
	ltostr(buf, con->request.content_length);
	fcgi_env_add(p->fcgi_env, CONST_STRING("CONTENT_LENGTH"), buf);
	
	
	if (host->docroot->used) {
		/* 
		 * rewrite SCRIPT_FILENAME 
		 */
		
		buffer_copy_string_buffer(p->path, host->docroot);
		BUFFER_APPEND_SLASH(p->path);
		buffer_append_string_buffer(p->path, con->uri.path);
		
		fcgi_env_add(p->fcgi_env, CONST_STRING("SCRIPT_FILENAME"), p->path->ptr);
		fcgi_env_add(p->fcgi_env, CONST_STRING("PATH_TRANSLATED"), p->path->ptr);
		fcgi_env_add(p->fcgi_env, CONST_STRING("DOCUMENT_ROOT"), host->docroot->ptr);
	} else {
		if (con->request.pathinfo->used) {
			fcgi_env_add(p->fcgi_env, CONST_STRING("PATH_INFO"), con->request.pathinfo->ptr);
		}
		
		fcgi_env_add(p->fcgi_env, CONST_STRING("SCRIPT_FILENAME"), con->physical.path->ptr);
		fcgi_env_add(p->fcgi_env, CONST_STRING("PATH_TRANSLATED"), con->physical.path->ptr);
		fcgi_env_add(p->fcgi_env, CONST_STRING("DOCUMENT_ROOT"), con->physical.doc_root->ptr);
	}
	fcgi_env_add(p->fcgi_env, CONST_STRING("SCRIPT_NAME"), con->uri.path->ptr);
	fcgi_env_add(p->fcgi_env, CONST_STRING("REQUEST_URI"), con->request.uri->ptr);
	fcgi_env_add(p->fcgi_env, CONST_STRING("QUERY_STRING"), con->uri.query->used ? con->uri.query->ptr : "");
	fcgi_env_add(p->fcgi_env, CONST_STRING("REQUEST_METHOD"), get_http_method_name(con->request.http_method));
	fcgi_env_add(p->fcgi_env, CONST_STRING("REDIRECT_STATUS"), "200");
	fcgi_env_add(p->fcgi_env, CONST_STRING("SERVER_PROTOCOL"), get_http_version_name(con->request.http_version));
	
	fcgi_env_add_request_headers(srv, con, p_d);
	
	fcgi_header(&(header), FCGI_PARAMS, request_id, p->fcgi_env->used, 0);
	buffer_append_memory(con->fcgi.write_buffer, (const char *)&header, sizeof(FCGI_Header));
	buffer_append_memory(con->fcgi.write_buffer, (const char *)p->fcgi_env->ptr, p->fcgi_env->used);
	
	fcgi_header(&(header), FCGI_PARAMS, request_id, 0, 0);
	buffer_append_memory(con->fcgi.write_buffer, (const char *)&header, sizeof(FCGI_Header));
	
	/* send FCGI_STDIN */
	
	/* something to send ? */
	for (offset = 0; offset != con->request.content_length; ) {
		/* send chunks of 1024 bytes */
		size_t toWrite = con->request.content_length - offset > 4096 ? 4096 : con->request.content_length - offset;
		
		fcgi_header(&(header), FCGI_STDIN, request_id, toWrite, 0);
		buffer_append_memory(con->fcgi.write_buffer, (const char *)&header, sizeof(FCGI_Header));
		buffer_append_memory(con->fcgi.write_buffer, (const char *)(con->request.content->ptr + offset), toWrite);
		
		offset += toWrite;
	}
	
	/* terminate STDIN */
	fcgi_header(&(header), FCGI_STDIN, request_id, 0, 0);
	buffer_append_memory(con->fcgi.write_buffer, (const char *)&header, sizeof(FCGI_Header));

#if 0
	for (i = 0; i < con->fcgi.write_buffer->used; i++) {
		fprintf(stderr, "%02x ", con->fcgi.write_buffer->ptr[i]);
		if ((i+1) % 16 == 0) {
			size_t j;
			for (j = i-15; j <= i; j++) {
				fprintf(stderr, "%c", 
					isprint((unsigned char)con->fcgi.write_buffer->ptr[j]) ? con->fcgi.write_buffer->ptr[j] : '.');
			}
			fprintf(stderr, "\n");
		}
	}
#endif
	
	return 0;
}

static int fcgi_set_state(server *srv, connection *con, fcgi_connection_state_t state) {
	con->fcgi.state = state;
	con->fcgi.state_timestamp = srv->cur_ts;
	
	return 0;
}

static int fcgi_write_request(server *srv, connection *con, void *p_d, data_fastcgi *host) {
	plugin_data *p = p_d;
	int r;
	
	if (!host || !host->host->used | !host->port) return -1;
	
	switch(con->fcgi.state) {
	case FCGI_STATE_INIT:
		if (-1 == (con->fcgi.fd = socket(AF_INET, SOCK_STREAM, 0))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "socket failed: ", strerror(errno));
			return -1;
		}
		con->fcgi.fde_ndx = -1;
		
		fdevent_fdc_add(srv->ev->fdc, con->fcgi.fd, con, FD_EVENT_TYPE_FCGI_CONNECTION);
		
		if (-1 == fdevent_fcntl_set(srv->ev, con->fcgi.fd)) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed: ", strerror(errno));
			
			con->fcgi.fde_ndx = fdevent_event_del(srv->ev, con->fcgi.fde_ndx, con->fcgi.fd);
			fdevent_fdc_del(srv->ev->fdc, con->fcgi.fd);
			close(con->fcgi.fd);
			return -1;
		}
		
		fcgi_set_state(srv, con, FCGI_STATE_CONNECT);
		/* fall through */
		
	case FCGI_STATE_CONNECT:
		/* try to finish the connect() */
		con->fcgi.fd = fcgi_establish_connection(srv, con, p_d, con->fcgi.host, con->fcgi.fd);
		
		if (con->fcgi.fd == -1) {
			con->fcgi.fde_ndx = -1;
			return -1;
		}
		
		if (con->fcgi.fd < 0) {
			con->fcgi.fd = -con->fcgi.fd;
			fcgi_set_state(srv, con, FCGI_STATE_CONNECT);
			return 0;
		}
		
		if (con->fcgi.request_id == 0) {
			con->fcgi.request_id = fcgi_requestid_new(srv, p);
		} else {
			log_error_write(srv, __FILE__, __LINE__, "sd", "fcgi-request is already in use:", con->fcgi.request_id);
		}
		
		fcgi_set_state(srv, con, FCGI_STATE_PREPARE_WRITE);
		/* fall through */
	case FCGI_STATE_PREPARE_WRITE:
		fcgi_create_env(srv, con, p_d, host, con->fcgi.request_id);
		
		fcgi_set_state(srv, con, FCGI_STATE_WRITE);
		con->fcgi.write_offset = 0;
		
		/* fall through */
	case FCGI_STATE_WRITE:
		/* continue with the code after the switch */
		if (-1 == (r = write(con->fcgi.fd, 
				     con->fcgi.write_buffer->ptr + con->fcgi.write_offset, 
				     con->fcgi.write_buffer->used - con->fcgi.write_offset))) {
			if (errno != EAGAIN) {
				log_error_write(srv, __FILE__, __LINE__, "ssd", "write failed:", strerror(errno), r);
				
				return -1;
			} else {
				return 0;
			}
		}
		
		con->fcgi.write_offset += r;
		
		if (con->fcgi.write_offset == con->fcgi.write_buffer->used) {
			fcgi_set_state(srv, con, FCGI_STATE_READ);
		}
		
		break;
	case FCGI_STATE_READ:
		/* waiting for a response */
		break;
	default:
		return -1;
	}
	
	return 0;
}

static int fcgi_response_parse(server *srv, connection *con, plugin_data *p, buffer *in) {
	char *s, *ns;
	
	UNUSED(srv);

	/* \r\n -> \0\0 */
	
	buffer_copy_string_buffer(p->parse_response, in);
	
	for (s = p->parse_response->ptr; NULL != (ns = strstr(s, "\r\n")); s = ns + 2) {
		char *key, *value;
		int key_len;
		data_string *ds;
		
		ns[0] = '\0';
		ns[1] = '\0';
		
		key = s;
		if (NULL == (value = strchr(s, ':'))) {
			/* we expect: "<key>: <value>\n" */
			continue;
		}
		
		key_len = value - key;
		
		value++;
		/* strip WS */
		while (*value == ' ' || *value == '\t') value++;
		
		if (NULL == (ds = (data_string *)array_get_unused_element(con->response.headers, TYPE_STRING))) {
			ds = data_string_init();
		}
		buffer_copy_string_len(ds->key, key, key_len);
		buffer_copy_string(ds->value, value);
		
		array_insert_unique(con->response.headers, (data_unset *)ds);
		
		switch(key_len) {
		case 4:
			if (0 == strncasecmp(key, "Date", key_len)) {
				con->parsed_response |= HTTP_DATE;
			}
			break;
		case 6:
			if (0 == strncasecmp(key, "Status", key_len)) {
				con->http_status = strtol(value, NULL, 10);
				con->parsed_response |= HTTP_STATUS;
			}
			break;
		case 8:
			if (0 == strncasecmp(key, "Location", key_len)) {
				con->parsed_response |= HTTP_LOCATION;
			}
			break;
		case 10:
			if (0 == strncasecmp(key, "Connection", key_len)) {
				con->response.keep_alive = (0 == strcasecmp(value, "Keep-Alive")) ? TRUE : FALSE;
				con->parsed_response |= HTTP_CONNECTION;
			}
			break;
		case 14:
			if (0 == strncasecmp(key, "Content-Length", key_len)) {
				con->response.content_length = strtol(value, NULL, 10);
				con->parsed_response |= HTTP_CONTENT_LENGTH;
			}
			break;
		default:
			break;
		}
	}
	
	/* CGI/1.1 rev 03 - 7.2.1.2 */
	if ((con->parsed_response & HTTP_LOCATION) &&
	    !(con->parsed_response & HTTP_STATUS)) {
		con->http_status = 302;
	}
	
	return 0;
}

static int fcgi_demux_response(server *srv, connection *con, void *p_d) {
	plugin_data *p = p_d;
	int fcgi_fd;
	ssize_t len;
	int fin = 0;
	int b;
	ssize_t r;
			
	fcgi_fd = con->fcgi.fd;
	
	/* check how much we have to read */
	if (ioctl(con->fcgi.fd, FIONREAD, &b)) {
		log_error_write(srv, __FILE__, __LINE__, "sd", "unexpected end-of-file (perhaps the fastcgi process died): ",
				fcgi_fd
				);
		return -1;
	}
	
	if (con->fcgi.rb == NULL) {
		con->fcgi.rb = malloc(sizeof(read_buffer));
		memset(con->fcgi.rb, 0, sizeof(read_buffer));
	}
	
	if (b > 0) {
		if (con->fcgi.rb->size == 0) {
			con->fcgi.rb->size = b;
			con->fcgi.rb->ptr = malloc(con->fcgi.rb->size * sizeof(char));
		} else if (con->fcgi.rb->size < con->fcgi.rb->used + b) {
			con->fcgi.rb->size += b;
			con->fcgi.rb->ptr = realloc(con->fcgi.rb->ptr, con->fcgi.rb->size * sizeof(char));
		}
		
		if (-1 == (r = read(con->fcgi.fd, con->fcgi.rb->ptr + con->fcgi.rb->used, b))) {
			log_error_write(srv, __FILE__, __LINE__, "sds", 
					"unexpected end-of-file (perhaps the fastcgi process died):",
					fcgi_fd, strerror(errno));
			return -1;
		}
		
		/* this should be catched by the b > 0 above */
		assert(r);
		
		con->fcgi.rb->used += r;
	} else {
		con->fcgi.fde_ndx = fdevent_event_del(srv->ev, con->fcgi.fde_ndx, con->fcgi.fd);
				fdevent_fdc_del(srv->ev->fdc, con->fcgi.fd);
				fcgi_requestid_del(srv, p, con->fcgi.request_id);
				close(con->fcgi.fd);
				con->fcgi.fd = -1;
				con->fcgi.fde_ndx = -1;
		
		log_error_write(srv, __FILE__, __LINE__, "sd", "unexpected end-of-file (perhaps the fastcgi process died): ",
				fcgi_fd
				);
		
		return -1;
	}
	
	/* read until it would block */
	while (fin == 0) {
		size_t request_id;
		
		if (con->fcgi.response_len == 0) {
			FCGI_Header *header;
			
			if (con->fcgi.rb->used - con->fcgi.rb->offset < sizeof(header)) {
#if 0
				log_error_write(srv, __FILE__, __LINE__, "sddd", "didn't get the full header: ",
						con->fcgi.rb->used - con->fcgi.rb->offset, sizeof(header),
						fcgi_fd
						);
#endif
				break;
			}
#if 0
			fprintf(stderr, "fcgi-version: %02x\n", con->fcgi.rb->ptr[con->fcgi.rb->offset]);
#endif
			
			header = (FCGI_Header *)(con->fcgi.rb->ptr + con->fcgi.rb->offset);
			con->fcgi.rb->offset += sizeof(FCGI_Header);
			
			len = (header->contentLengthB0 | (header->contentLengthB1 << 8)) + header->paddingLength;
			request_id = (header->requestIdB0 | (header->requestIdB1 << 8));

			con->fcgi.response_len = len;
			con->fcgi.response_request_id = request_id;
			con->fcgi.response_type = header->type;
			con->fcgi.response_padding = header->paddingLength;
			
#if 0
			log_error_write(srv, __FILE__, __LINE__, "sddd", "offset: ",
					fcgi_fd, con->fcgi.rb->offset, header->type
					);
#endif
			
		} else {
			len = con->fcgi.response_len;
		}
		
		if (con->fcgi.rb->used - con->fcgi.rb->offset < con->fcgi.response_len) {
			/* we are not finished yet */
			break;
		}
		
		con->fcgi.response->ptr = con->fcgi.rb->ptr + con->fcgi.rb->offset;
		con->fcgi.rb->offset += con->fcgi.response_len;
#if 0
		log_error_write(srv, __FILE__, __LINE__, "sdd", "offset: ",
				fcgi_fd, con->fcgi.rb->offset
				);
#endif
		
		/* remove padding */
#if 0
		con->fcgi.response->ptr[con->fcgi.response_len - con->fcgi.response_padding] = '\0';
#endif
		con->fcgi.response->used = con->fcgi.response_len - con->fcgi.response_padding + 1;
		
		/* mark the fast-cgi packet as finished */
		con->fcgi.response_len = 0;
		
		switch(con->fcgi.response_type) {
		case FCGI_STDOUT:
			if (len) {
#if 0
				log_error_write(srv, __FILE__, __LINE__, "sds", "len", len, con->fcgi.response->ptr);
#endif
				
				if (0 == con->got_response) {
					con->got_response = 1;
					buffer_prepare_copy(con->fcgi.response_header, 128);
				}
				
				if (0 == con->file_started) {
					char *c;
					
					/* search for the \r\n\r\n in the string */
					if (NULL != (c = buffer_search_string_len(con->fcgi.response, "\r\n\r\n", 4))) {
						size_t hlen = c - con->fcgi.response->ptr + 4;
						size_t blen = con->fcgi.response->used - hlen - 1;
						/* found */
						
						buffer_append_string_len(con->fcgi.response_header, con->fcgi.response->ptr, c - con->fcgi.response->ptr + 4);
#if 0
						log_error_write(srv, __FILE__, __LINE__, "ss", "Header:", con->fcgi.response_header->ptr);
#endif
						/* parse the response header */
						fcgi_response_parse(srv, con, p, con->fcgi.response_header);
						
						/* enable chunked-transfer-encoding */
						if (con->request.http_version == HTTP_VERSION_1_1 &&
						    !(con->parsed_response & HTTP_CONTENT_LENGTH)) {
							con->response.transfer_encoding = HTTP_TRANSFER_ENCODING_CHUNKED;
						}

						con->file_started = 1;
						
						if (blen) {
							http_chunk_append_mem(srv, con, c + 4, blen + 1);
							joblist_append(srv, con);
#if 0
							log_error_write(srv, __FILE__, __LINE__, "sd", "body-len", blen);
#endif
						}
					} else {
						/* copy everything */
						buffer_append_string(con->fcgi.response_header, con->fcgi.response->ptr);
					}
				} else {
					http_chunk_append_mem(srv, con, con->fcgi.response->ptr, con->fcgi.response->used);
					joblist_append(srv, con);
#if 0
					log_error_write(srv, __FILE__, __LINE__, "sd", "body-len", con->fcgi.response->used);
#endif
				}
			} else {
				/* finished */
			}
			break;
		case FCGI_END_REQUEST:
			con->file_finished = 1;
			
			http_chunk_append_mem(srv, con, NULL, 0);
			joblist_append(srv, con);
			
			con->fcgi.fde_ndx = fdevent_event_del(srv->ev, con->fcgi.fde_ndx, con->fcgi.fd);
			fdevent_fdc_del(srv->ev->fdc, con->fcgi.fd);
			fcgi_requestid_del(srv, p, con->fcgi.request_id);
			close(con->fcgi.fd);
			con->fcgi.fd = -1;
			con->fcgi.fde_ndx = -1;
			
			con->fcgi.rb->used = 0;
			con->fcgi.rb->offset = 0;
			
			fin = 1;
			break;
		default:
			log_error_write(srv, __FILE__, __LINE__, "sd", "fcgi: header.type not handled: ", con->fcgi.response_type);
			break;
		}
	}
	
	con->fcgi.response->ptr = NULL;

	return fin;
}


static handler_t fcgi_is_handled(server *srv, connection *con, void *p_d) {
	plugin_data *p = p_d;
	size_t s_len;
	int used = -1;
	int ndx;
	size_t k;
	buffer *fn = con->physical.path;
	data_array *extension = NULL;
	
	if (fn->used == 0) return HANDLER_ERROR;
	
	s_len = fn->used - 1;
	
	/* check if extension matches */
	for (k = 0; k < p->extensions->used; k++) {
		size_t ct_len;
		
		extension = (data_array *)p->extensions->data[k];
		
		if (extension->key->used == 0) continue;
		
		ct_len = extension->key->used - 1;
		
		if (s_len < ct_len) continue;
		
		if (0 == strncmp(fn->ptr + s_len - ct_len, extension->key->ptr, ct_len)) {
			break;
		}
	}
	
	/* extension doesn't match */
	if (k == p->extensions->used) return HANDLER_GO_ON;
		
	/* get best server */
	for (k = 0, ndx = -1; k < extension->value->used; k++) {
		data_fastcgi *host = (data_fastcgi *)extension->value->data[k];
		
		/* enable the server again, perhaps it is back again */
		if ((host->usage == -1) &&
		    (srv->cur_ts - host->disable_ts > FCGI_RETRY_TIMEOUT)) {
			host->usage = 0;
			
			log_error_write(srv, __FILE__, __LINE__,  "sbd", "fcgi-server re-enabled:", 
					host->host, host->port);
		}
		
		if (used == -1 || host->usage < used) {
			used = host->usage;
			
			ndx = k;
		}
	}
	
	/* found a server */
	if (ndx != -1) {
		data_fastcgi *host = (data_fastcgi *)extension->value->data[ndx];
		
		host->usage++;
		
		con->fcgi.host = host;
		
		/* ok, create the request */
		if (-1 == (fcgi_write_request(srv, con, p_d, host))) {
			log_error_write(srv, __FILE__, __LINE__,  "sbdd", "fcgi-server disabled:", 
					host->host,
					host->port,
					con->fcgi.fd);
			
			/* disable this server */
			host->usage = -1;
			host->disable_ts = srv->cur_ts;
			
			con->mode = DIRECT;
			con->http_status = 503;
			return HANDLER_FINISHED;
		}
		
		con->mode = FASTCGI;
		
		if (con->file_started) {
			return HANDLER_FINISHED;
		} else {
			return HANDLER_WAIT_FOR_EVENT;
		}
	} else {
		/* no handler found */
		con->http_status = 500;
		
		log_error_write(srv, __FILE__, __LINE__,  "sb", 
				"no fcgi-handler found for:", 
				fn);
		
		return HANDLER_FINISHED;
	}
	
	return HANDLER_WAIT_FOR_EVENT;
}

static handler_t fcgi_connection_close(server *srv, connection *con, void *p_d) {
	plugin_data *p = p_d;
	
	if (con->mode != FASTCGI) return HANDLER_GO_ON;
	
	log_error_write(srv, __FILE__, __LINE__, "ssdsd", 
			"emergency exit: fastcgi:", 
			"connection-fd:", con->fd,
			"fcgi-fd:", con->fcgi.fd);
	
	con->fcgi.fde_ndx = fdevent_event_del(srv->ev, con->fcgi.fde_ndx, con->fcgi.fd);
	fdevent_fdc_del(srv->ev->fdc, con->fcgi.fd);
	fcgi_requestid_del(srv, p, con->fcgi.request_id);
	close(con->fcgi.fd);
	con->fcgi.fd = -1;
	con->fcgi.fde_ndx = -1;
	
	return HANDLER_FINISHED;
}

SUBREQUEST_FUNC(mod_fastcgi_handle_subrequest) {
	data_fastcgi *host = con->fcgi.host;
	
	/* not my job */
	if (con->mode != FASTCGI) return HANDLER_GO_ON;
	
	/* ok, create the request */
	if (-1 == fcgi_write_request(srv, con, p_d, host)) {
		log_error_write(srv, __FILE__, __LINE__,  "sbdd", "fcgi-server disabled:", 
				host->host,
				host->port,
				con->fcgi.fd);
		
		/* disable this server */
		host->usage = -1;
		host->disable_ts = srv->cur_ts;
		
		con->mode = DIRECT;
		con->http_status = 503;
		return HANDLER_FINISHED;
	}
	
	if (con->file_started) {
		return HANDLER_FINISHED;
	} else {
		return HANDLER_WAIT_FOR_EVENT;
	}
}

static handler_t fcgi_handle_fdevent(server *srv, const fd_conn *fdc, void *p_d) {
	plugin_data *p = p_d;
	connection *con = fdc->conn;
	
	if (fdc->fd_type != FD_EVENT_TYPE_FCGI_CONNECTION) return HANDLER_GO_ON;
	
	joblist_append(srv, con);
	
	if ((fdc->revents & FDEVENT_IN) &&
	    con->fcgi.state == FCGI_STATE_READ) {
		switch (fcgi_demux_response(srv, con, p)) {
		case 0:
			break;
		case 1:
			con->fcgi.host->usage--;
			
			break;
		case -1:
			if (con->file_started == 0) {
				/* nothing has been send out yet, send a 500 */
				connection_set_state(srv, con, HANDLE_REQUEST);
				con->http_status = 500;
				con->mode = DIRECT;
			} else {
				/* response might have been already started, kill the connection */
				connection_set_state(srv, con, ERROR);
			}
			
			return HANDLER_FINISHED;
		}
	}
	
	if (fdc->revents & FDEVENT_OUT) {
		/* nothing to do */
		if (con->fcgi.state == FCGI_STATE_CONNECT ||
		    con->fcgi.state == FCGI_STATE_WRITE) {
			
			return mod_fastcgi_handle_subrequest(srv, con, p_d);
		} else {
			log_error_write(srv, __FILE__, __LINE__, "sd", "fcgi: out", con->fcgi.state);
		}
	}
	
	/* perhaps this issue is already handled */
	if (fdc->revents & FDEVENT_HUP) {
		log_error_write(srv, __FILE__, __LINE__, "sbSBSDS", 
				"error: unexpected close of fastcgi connection for", 
				con->uri.path,
				"(no fastcgi process on host: ", 
				con->fcgi.host->host,
				", port: ", 
				con->fcgi.host->port,
				" ?)" );
		
#ifndef USE_LINUX_SIGIO
		fcgi_connection_close(srv, con, p);
# if 0
		log_error_write(srv, __FILE__, __LINE__, "sd", "fcgi-FDEVENT_HUP", con->fd);
# endif			
		return HANDLER_ERROR;
#endif
	} else if (fdc->revents & FDEVENT_ERR) {
		log_error_write(srv, __FILE__, __LINE__, "s", "fcgi: err");
		/* kill all connections to the fastcgi process */
		
		fcgi_connection_close(srv, con, p);
#if 1
		log_error_write(srv, __FILE__, __LINE__, "s", "fcgi-FDEVENT_ERR");
#endif			
		return HANDLER_ERROR;
	}
	
	return HANDLER_FINISHED;
}

JOBLIST_FUNC(mod_fastcgi_handle_joblist) {
	fd_conn_buffer *fdc = srv->ev->fdc;

	UNUSED(p_d);

	if (con->fcgi.fd != -1 &&
	    (con->fcgi.fde_ndx == -1 || 
	     fdc->ptr[con->fcgi.fd].fd_type == FD_EVENT_TYPE_FCGI_CONNECTION)) {
		switch (con->fcgi.state) {
		case FCGI_STATE_READ:
			con->fcgi.fde_ndx = fdevent_event_add(srv->ev, con->fcgi.fde_ndx, con->fcgi.fd, FDEVENT_IN);
			
			break;
		case FCGI_STATE_CONNECT:
			con->fcgi.fde_ndx = fdevent_event_add(srv->ev, con->fcgi.fde_ndx, con->fcgi.fd, FDEVENT_OUT);
			
			break;
		default:
			log_error_write(srv, __FILE__, __LINE__, "sd", "unhandled fcgi.state", con->fcgi.state);
			break;
		}
	}

	return HANDLER_GO_ON;
}



int mod_fastcgi_plugin_init(plugin *p) {
	p->name        = buffer_init_string("fastcgi");

	p->init     = mod_fastcgi_init;
	p->cleanup     = mod_fastcgi_free;
	p->set_defaults = mod_fastcgi_set_defaults;
	p->handle_connection_close = fcgi_connection_close;
	p->handle_subrequest_start = fcgi_is_handled;
	p->handle_subrequest = mod_fastcgi_handle_subrequest;
	p->handle_fdevent = fcgi_handle_fdevent;
	p->handle_joblist = mod_fastcgi_handle_joblist;
	
	p->data        = NULL;
	
	return 0;
}
