#include <sys/types.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#define __USE_GNU
#include <fcntl.h>

#include "fdevent.h"
#include "settings.h"

#ifdef USE_LINUX_SIGIO
static void fdevent_linux_rtsig_free(fdevents *ev) {
	free(ev->pollfds);
	free(ev->unused.ptr);
}


static int fdevent_linux_rtsig_event_del(fdevents *ev, int fde_ndx, int fd) {
	if (fde_ndx < 0) return -1;
	
	if ((size_t)fde_ndx + 1 > ev->used) {
		fprintf(stderr, "%s.%d: del! out of range %d %u\n", __FILE__, __LINE__, fde_ndx, ev->used);
		SEGFAULT();
	}
	
	if (ev->pollfds[fde_ndx].fd == fd) {
		size_t k = fde_ndx;
		
		ev->pollfds[k].fd = -1;
		/* ev->pollfds[k].events = 0; */
		/* ev->pollfds[k].revents = 0; */
		
		if (ev->unused.size == 0) {
			ev->unused.size = 16;
			ev->unused.ptr = malloc(sizeof(fd_unused) * ev->unused.size);
		} else if (ev->unused.size == ev->unused.used) {
			ev->unused.size += 16;
			ev->unused.ptr = realloc(ev->unused.ptr, sizeof(fd_unused) * ev->unused.size);
		}
		
		ev->unused.ptr[ev->unused.used].ndx     = k;
		
		ev->unused.used++;
	} else {
		fprintf(stderr, "%s.%d: del! %d %d\n", __FILE__, __LINE__, ev->pollfds[fde_ndx].fd, fd);
		
		SEGFAULT();
	}
	
	return -1;
}

static int fdevent_linux_rtsig_event_compress(fdevents *ev) {
	size_t j;
	int i;
	
	if (ev->unused.used == 0) return 0;
	
	for (j = 0; j < ev->unused.used; j++) {
		size_t k = ev->unused.ptr[j].ndx;
		size_t good_ndx;

		for (good_ndx = ev->used - 1; good_ndx + 1 > 0 && ev->pollfds[good_ndx].fd == -1; good_ndx--);
		
		if (good_ndx + 1 == 0) {
			SEGFAULT();
		}
		
		if (k < good_ndx) {
			connection       *con;
			file_cache_entry *fce;
			
			/* perhaps this section is a problem if the connection is gone while we have link to it */
			
			ev->pollfds[k].fd      = ev->pollfds[good_ndx].fd;
			ev->pollfds[k].events  = ev->pollfds[good_ndx].events;
			ev->pollfds[k].revents = ev->pollfds[good_ndx].revents;
			
			ev->pollfds[good_ndx].fd = -1;

			if (-1 == (i = fdevent_fdc_get_index(ev->fdc, ev->pollfds[k].fd))) {
				SEGFAULT();
			}
			
			con = ev->fdc->ptr[i].conn;
			fce = ev->fdc->ptr[i].conn;
				
			switch (ev->fdc->ptr[i].fd_type) {
			case FD_EVENT_TYPE_CONNECTION:
				if (con->fde_ndx != -1) con->fde_ndx = k;
				break;
			case FD_EVENT_TYPE_FCGI_CONNECTION:
				if (con->fcgi.fde_ndx != -1) con->fcgi.fde_ndx = k;
				break;
			case FD_EVENT_TYPE_CGI_CONNECTION:
				if (con->cgi.fde_ndx != -1) con->cgi.fde_ndx = k;
				break;
			case FD_EVENT_TYPE_DIRWATCH:
				if (fce->fde_ndx != -1) fce->fde_ndx = k;
				break;
			case -1:
				break;
			default:
				fprintf(stderr, "%s.%d: compress: (%d)\n", __FILE__, __LINE__, ev->pollfds[k].fd);
				SEGFAULT();
				break;
			}
		}
		
		ev->used--;
	}

	ev->unused.used = 0;
	
	return 0;
}

static int fdevent_linux_rtsig_event_add(fdevents *ev, int fde_ndx, int fd, int events) {
	/* known index */
	if (fde_ndx != -1) {
		if (ev->pollfds[fde_ndx].fd == fd) {
			ev->pollfds[fde_ndx].events = events;
			
			return fde_ndx;
		}
		fprintf(stderr, "%s.%d: add: (%d, %d)\n", __FILE__, __LINE__, fde_ndx, ev->pollfds[fde_ndx].fd);
		SEGFAULT();
	}
	
	if (ev->size == 0) {
		ev->size = 16;
		ev->pollfds = malloc(sizeof(struct pollfd) * ev->size);
	} else if (ev->size == ev->used) {
		ev->size += 16;
		ev->pollfds = realloc(ev->pollfds, sizeof(struct pollfd) * ev->size);
	}
	
	ev->pollfds[ev->used].fd = fd;
	ev->pollfds[ev->used].events = events;
	
	return ev->used++;
}

static int fdevent_linux_rtsig_poll(fdevents *ev, int timeout_ms) {
	struct timespec ts;
	int r;
	
	fdevent_linux_rtsig_event_compress(ev);
	
	ev->in_sigio = 1;
		
	ts.tv_sec =  timeout_ms / 1000;
	ts.tv_nsec = (timeout_ms * 1000000) % 1000000000;
	r = sigtimedwait(&(ev->sigset), &(ev->siginfo), &(ts));
		
	if (r == -1) { 
		if (errno == EAGAIN) return 0;
		return r; 
	} else if (r == SIGIO) {
		ev->in_sigio = 0;
		
		r = poll(ev->pollfds, ev->used, timeout_ms);
		
		return r;
	} else if (r == ev->signum) {
#  if 0
		fprintf(stderr, "event: %d %02lx\n", ev->siginfo.si_fd, ev->siginfo.si_band);
#  endif
		return 1;
	} else {
		/* ? */
		return -1;
	}
}

static int fdevent_linux_rtsig_event_get_revent(fdevents *ev, size_t ndx) {
	if (ev->in_sigio == 1) {
#  if 0
		if (ev->siginfo.si_band == POLLERR) {
			fprintf(stderr, "event: %d %02lx %02x %s\n", ev->siginfo.si_fd, ev->siginfo.si_band, errno, strerror(errno));
		}
#  endif		
		if (ndx != 0) {
			fprintf(stderr, "+\n");
			return 0;
		}
		
		return ev->siginfo.si_band & 0x3f;
	} else {
		if (ndx >= ev->used) {
			fprintf(stderr, "%s.%d: event: %u %u\n", __FILE__, __LINE__, ndx, ev->used);
			return 0;
		}
		return ev->pollfds[ndx].revents;
	}
}

static int fdevent_linux_rtsig_event_get_fd(fdevents *ev, size_t ndx) {
	if (ev->in_sigio == 1) {
		return ev->siginfo.si_fd;
	} else {
		return ev->pollfds[ndx].fd;
	}
}

static int fdevent_linux_rtsig_fcntl_set(fdevents *ev, int fd) {
	static pid_t pid = 0;
	
	if (pid == 0) pid = getpid();
	
	if (-1 == fcntl(fd, F_SETSIG, ev->signum)) return -1;
	
	if (-1 == fcntl(fd, F_SETOWN, (int) pid)) return -1;
	
	return fcntl(fd, F_SETFL, O_ASYNC | O_NONBLOCK | O_RDWR);
}


static int fdevent_linux_rtsig_event_next_fdndx(fdevents *ev, int ndx) {
	if (ev->in_sigio == 1) {
		if (ndx < 0) return 0;
		return -1;
	} else {
		size_t i;
		
		i = (ndx < 0) ? 0 : ndx + 1;
		for (; i < ev->used; i++) {
			if (ev->pollfds[i].revents) break;
		}
		
		if (i > INT_MAX) {
			SEGFAULT();
		}
		
		return i;
	}
}

int fdevent_linux_rtsig_init(fdevents *ev, size_t maxfds) {
	UNUSED(maxfds);

	ev->type = FDEVENT_HANDLER_LINUX_RTSIG;
#define SET(x) \
	ev->x = fdevent_linux_rtsig_##x;
	
	SET(free);
	SET(poll);
	
	SET(event_del);
	SET(event_add);
	
	SET(event_next_fdndx);
	SET(fcntl_set);
	SET(event_get_fd);
	SET(event_get_revent);
	
	ev->signum = SIGRTMIN + 1;
	
	sigemptyset(&(ev->sigset));
	sigaddset(&(ev->sigset), ev->signum);
	sigaddset(&(ev->sigset), SIGIO);
	if (-1 == sigprocmask(SIG_BLOCK, &(ev->sigset), NULL)) {
		fprintf(stderr, "%s.%d: sigprocmask failed (%s), try to set server.event-handler = \"poll\" or \"select\" \n",
			__FILE__, __LINE__, strerror(errno));
		
		return -1;
	}
	
	ev->in_sigio = 1;
	
	return 0;
}
#else
int fdevent_linux_rtsig_init(fdevents *ev, size_t maxfds) {
	UNUSED(ev);
	UNUSED(maxfds);

	fprintf(stderr, "%s.%d: linux-rtsig not supported, try to set server.event-handler = \"poll\" or \"select\" \n",
		__FILE__, __LINE__);
	return -1;
}
#endif
