#include <sys/types.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>

#include "fdevent.h"
#include "settings.h"

#ifdef USE_SOLARIS_DEVPOLL

#define FDEVENT_SOLARIS_MAX_FDS 128

static void fdevent_solaris_devpoll_free(fdevents *ev) {
	free(ev->devpollfds);
	close(ev->devpoll_fd);
}

/* return -1 is fine here */

static int fdevent_solaris_devpoll_event_del(fdevents *ev, int fde_ndx, int fd) {
	struct pollfd pfd;
		
	if (fde_ndx == -1) return -1;
	
	pfd.fd = fd;
	pfd.events = POLLREMOVE;
	pfd.revents = 0;
	
	if (-1 == write(ev->devpoll_fd, &pfd, sizeof(struct pollfd))) {
		fprintf(stderr, "%s.%d: (del) write failed: (%d, %s)\n", 
			__FILE__, __LINE__, 
			fd, strerror(errno));
		
		return -1;
	}
	
	return -1;
}

static int fdevent_solaris_devpoll_event_add(fdevents *ev, int fde_ndx, int fd, int events) {
	struct pollfd pfd;
	int add = 0;
		
	if (fde_ndx == -1) add = 1;
	
	pfd.fd = fd;
	pfd.events = events;
	pfd.revents = 0;
	
	if (-1 == write(ev->devpoll_fd, &pfd, sizeof(struct pollfd))) {
		fprintf(stderr, "%s.%d: (del) write failed: (%d, %s)\n", 
			__FILE__, __LINE__, 
			fd, strerror(errno));
		
		return -1;
	}
	
	return fd;
}

static int fdevent_solaris_devpoll_poll(fdevents *ev, int timeout_ms) {
	struct dvpoll dopoll;
	int ret;
	
	dopoll.dp_timeout = timeout_ms;
	dopoll.dp_nfds = FDEVENT_SOLARIS_MAX_FDS;
	dopoll.dp_fds = ev->devpollfds;
	
	ret = ioctl(ev->devpoll_fd, DP_POLL, &dopoll);
	
	return ret;
}

static int fdevent_solaris_devpoll_event_get_revent(fdevents *ev, size_t ndx) {
	return ev->devpollfds[ndx].revents;
}

static int fdevent_solaris_devpoll_event_get_fd(fdevents *ev, size_t ndx) {
	return ev->devpollfds[ndx].fd;
}

static int fdevent_solaris_devpoll_fcntl_set(fdevents *ev, int fd) {
	UNUSED(ev);

	return fcntl(fd, F_SETFL, O_NONBLOCK | O_RDWR);
}


static int fdevent_solaris_devpoll_event_next_fdndx(fdevents *ev, int last_ndx) {
	size_t i;
	
	UNUSED(ev);

	i = (last_ndx < 0) ? 0 : last_ndx + 1;
	
	return i;
}

int fdevent_solaris_devpoll_init(fdevents *ev, size_t maxfds) {
	UNUSED(maxfds);

	ev->type = FDEVENT_HANDLER_SOLARIS_DEVPOLL;
#define SET(x) \
	ev->x = fdevent_solaris_devpoll_##x;
	
	SET(free);
	SET(poll);
	
	SET(event_del);
	SET(event_add);
	
	SET(event_next_fdndx);
	SET(fcntl_set);
	SET(event_get_fd);
	SET(event_get_revent);
	
	ev->devpollfds = malloc(sizeof(struct pollfd) * FDEVENT_SOLARIS_MAX_FDS);
	
	if ((ev->devpoll_fd = open("/dev/poll", O_RDWR)) < 0) {
		fprintf(stderr, "%s.%d: opening /dev/poll failed (%s), try to set server.event-handler = \"poll\" or \"select\" \n",
			__FILE__, __LINE__, strerror(errno));
	}
	
	return 0;
}

#else
int fdevent_solaris_devpoll_init(fdevents *ev, size_t maxfds) {
	UNUSED(ev);
	UNUSED(maxfds);

	fprintf(stderr, "%s.%d: solaris-devpoll not supported, try to set server.event-handler = \"poll\" or \"select\" \n",
			__FILE__, __LINE__);
	
	return -1;
}
#endif
