#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>

#include "log.h"
#include "file_cache.h"
#include "fdevent.h"
#include "etag.h"

#ifndef O_LARGEFILE
# define O_LARGEFILE 0
#endif

#undef USE_LINUX_SIGIO

file_cache *file_cache_init(void) {
	file_cache *fc = NULL;
	
	fc = malloc(sizeof(file_cache));
	memset(fc, 0, sizeof(file_cache));
	
	return fc;
}

static file_cache_entry * file_cache_entry_init(void) {
	file_cache_entry *fce = NULL;
	
	fce = malloc(sizeof(file_cache_entry));
	memset(fce, 0, sizeof(file_cache_entry));
	
	fce->fd = -1;
	fce->fde_ndx = -1;
	fce->name = buffer_init();
	fce->etag = buffer_init();
	
	return fce;
}

static void file_cache_entry_free(file_cache_entry *fce) {
	if (!fce) return;
	
	buffer_free(fce->etag);
	buffer_free(fce->name);
	
	free(fce);
}

void file_cache_free(file_cache *fc) {
	size_t i;
	for (i = 0; i < fc->used; i++) {
		file_cache_entry_free(fc->ptr[i]);
	}
	
	free(fc->ptr);
	
	free(fc);
	
}

file_cache_entry * file_cache_get_entry(server *srv, connection *con, buffer *name, file_cache_entry *o_fce) {
	file_cache_entry *fce = NULL;
	file_cache *fc = srv->file_cache;
	size_t i;
	size_t oldest_ndx;
	
	UNUSED(con);

	/* still valid ? */
	if (o_fce != NULL) {
		if (buffer_is_equal(name, o_fce->name) &&
		    (o_fce->fd != -1) &&
		    (o_fce->stat_ts == srv->cur_ts)
		    ) {
			return o_fce;
		} else {
			o_fce->in_use--;
		}
	}
	
	oldest_ndx = fc->used;
	
	/* check the cache */
	for (i = 0; i < fc->used; i++) {
		fce = fc->ptr[i];
		
		if (buffer_is_equal(name, fce->name)) {

#ifdef USE_LINUX_SIGIO
			if (fce->is_dirty == 0) {
				fce->in_use++;
				return fce;
			}
#endif
			
			if (-1 == (stat(name->ptr, &(fce->st)))) {
				int oerrno = errno;
				fce->in_use = 0;
				
				if (fce->fd >= 0) {
					close(fce->fd);
#ifdef USE_LINUX_SIGIO		
					/* unregister fd */
					if (fce->fde_ndx != -1) {
						log_error_write(srv, __FILE__, __LINE__, "sdds", "fdevent_event_del: ", fce->fde_ndx, fce->fd, fce->name->ptr);
						
						fce->fde_ndx = fdevent_event_del(srv->ev, fce->fde_ndx, fce->fd);
						fdevent_fdc_del(srv->ev->fdc, fce->fd);
					}
#endif

					fce->fd = -1;
				}
				errno = oerrno;
#if 0				
				log_error_write(srv, __FILE__, __LINE__, "sss", "stat failed: ", name->ptr, strerror(errno));
#endif				
				return NULL;
			}
			fce->stat_ts = srv->cur_ts;
			
			etag_create(srv->file_cache_etag, &(fce->st));
			
			if (!buffer_is_equal(srv->file_cache_etag, fce->etag)) {
				/* etag has changed: reopen file */
				if (fce->fd >= 0) close(fce->fd);
				if (-1 == (fce->fd = open(fce->name->ptr, O_RDONLY | O_LARGEFILE))) {
					log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
					
					buffer_reset(fce->name);
					return NULL;
				}
			}
			
			buffer_copy_string_buffer(fce->etag, srv->file_cache_etag);
			
			
#ifdef USE_LINUX_SIGIO
			fce->is_dirty = 0;
#endif
			
			fce->in_use++;
			return fce;
		}
		
		if (fce->in_use == 0 &&
		    ((oldest_ndx == fc->used) ||
		     (fce->stat_ts < fc->ptr[oldest_ndx]->stat_ts))) {
			oldest_ndx = i;
		}
	}
	
	/* not found, insert */
	if (fc->size == 0) {
		fc->size = 16;
		fc->ptr = malloc(sizeof(file_cache_entry *) * fc->size);
		memset(fc->ptr, 0, sizeof(file_cache_entry *) * fc->size);
		fc->used = 0;
	}
	
	if (fc->used < fc->size) {
		fc->ptr[fc->used] = file_cache_entry_init();
		fce = fc->ptr[fc->used++];
	} else if (oldest_ndx != fc->used) {
		/* the place is unused, reuse it */
		
		fce = fc->ptr[oldest_ndx];
		if (fce->fd >= 0) {
			close(fce->fd);
		
#ifdef USE_LINUX_SIGIO
			if (fce->fde_ndx != -1) {
# if 0
				log_error_write(srv, __FILE__, __LINE__, "sdds", "fdevent_event_del: ", fce->fde_ndx, fce->fd, fce->name->ptr);
# endif			
				fce->fde_ndx = fdevent_event_del(srv->ev, fce->fde_ndx, fce->fd);
				fdevent_fdc_del(srv->ev->fdc, fce->fd);
			}
#endif
		
			fce->fd = -1;
		}
	} else {
		/* the cache is full, time to resize */
		
		fc->size += 16;
		fc->ptr = realloc(fc->ptr, sizeof(file_cache_entry *) * fc->size);
		
		fc->ptr[fc->used] = file_cache_entry_init();
		fce = fc->ptr[fc->used++];
	}
	
	buffer_copy_string_buffer(fce->name, name);
	fce->in_use = 0;
	fce->fd = -1;
	
	if (-1 == stat(name->ptr, &(fce->st))) {
		int oerrno = errno;
#if 0
		log_error_write(srv, __FILE__, __LINE__, "sss", "stat failed: ", name->ptr, strerror(errno));
#endif
		
		if (fce->fd >= 0) {
			close(fce->fd);
		
#ifdef USE_LINUX_SIGIO	
			if (fce->fde_ndx != -1) {
# if 0
				log_error_write(srv, __FILE__, __LINE__, "sdds", "fdevent_event_del: ", fce->fde_ndx, fce->fd, fce->name->ptr);
# endif			
				fce->fde_ndx = fdevent_event_del(srv->ev, fce->fde_ndx, fce->fd);
				fdevent_fdc_del(srv->ev->fdc, fce->fd);
			}
#endif
		
			fce->fd = -1;
		}
		
		errno = oerrno;
		buffer_reset(fce->name);
		return NULL;
	}
	
	fce->stat_ts = srv->cur_ts;
	
	if (S_ISREG(fce->st.st_mode)) {
		if (-1 == (fce->fd = open(name->ptr, O_RDONLY | O_LARGEFILE))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
			
			buffer_reset(fce->name);
			return NULL;
		}
		
		etag_create(fce->etag, &(fce->st));
	} else if (S_ISDIR(fce->st.st_mode)) {
#ifdef USE_LINUX_SIGIO
		/* register sigio for the directory */
		
		if (-1 == (fce->fd = open(name->ptr, O_RDONLY))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
			buffer_reset(fce->name);
			
			return NULL;
		}
		
		if (fcntl(fce->fd, F_NOTIFY, DN_CREATE|DN_DELETE|DN_MODIFY|DN_MULTISHOT) < 0) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed: ", strerror(errno));
			buffer_reset(fce->name);
			return NULL;
		}
		/* register fd-handler */
		
		fdevent_fdc_add(srv->ev->fdc, fce->fd, fce, FD_EVENT_TYPE_DIRWATCH);
		fce->fde_ndx = fdevent_event_add(srv->ev, -1, fce->fd, FDEVENT_IN);
# if 0		
		log_error_write(srv, __FILE__, __LINE__, "sdds", "fdevent_event_add: ", fce->fde_ndx, fce->fd, fce->name->ptr);
# endif
#endif
	}
	
	fce->in_use++;
		
	return fce;
}

int file_cache_entry_release(server *srv, connection *con, file_cache_entry *fce) {
	UNUSED(srv);
	UNUSED(con);

	if (fce->in_use > 0) fce->in_use--;
	
	return 0;
}


int file_cache_handle_fdevent(server *srv, file_cache_entry *fce, int revent) {
	size_t i;
	file_cache *fc = srv->file_cache;;

	UNUSED(revent);
	/* */
#if 0	
	log_error_write(srv, __FILE__, __LINE__, "sds", "dir has changed: ", fce->fd, fce->name->ptr);
#endif	
	/* touch all files below this directory */
	
	for (i = 0; i < fc->used; i++) {
		file_cache_entry *f = fc->ptr[i];
		
		if (fce == f) continue;
		
		if (0 == strncmp(fce->name->ptr, f->name->ptr, fce->name->used - 1)) {
#if 0
			log_error_write(srv, __FILE__, __LINE__, "ss", "file hit: ", f->name->ptr);
#endif
			f->is_dirty = 1;
		}
	}
	
	return 0;
}
