#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include <stdio.h>

#include "base.h"
#include "log.h"
#include "buffer.h"

#include "plugin.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

typedef struct {
	buffer *access_log;
	buffer *ts_accesslog_str;
	
	int log_access_fd;
	time_t last_generated_accesslog_ts;
	
	buffer *conf_access_logfile;
} plugin_data;

INIT_FUNC(mod_accesslog_init) {
	plugin_data *p;
	
	p = malloc(sizeof(plugin_data));
	memset(p, 0, sizeof(plugin_data));
	
	p->access_log = buffer_init();
	p->ts_accesslog_str = buffer_init();
	p->conf_access_logfile = buffer_init();
	
	return p;
}

FREE_FUNC(mod_accesslog_free) {
	plugin_data *p = p_d;
	
	UNUSED(srv);

	if (!p) return HANDLER_GO_ON;
	
	if (p->access_log->used) {
		write(p->log_access_fd, p->access_log->ptr, p->access_log->used - 1);
	}
	
	if (p->log_access_fd != 1) close(p->log_access_fd);
	
	buffer_free(p->ts_accesslog_str);
	buffer_free(p->access_log);
	buffer_free(p->conf_access_logfile);
	
	free(p);
	
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC(log_access_open) {
	plugin_data *p = p_d;
	
	config_values_t cv[] = { 
		{ "accesslog.filename",             NULL, T_CONFIG_STRING },
		{ NULL,                             NULL, T_CONFIG_UNSET }
	};
	
	if (!p) return HANDLER_ERROR;
	cv[0].destination = p->conf_access_logfile;
	
	if (0 != config_insert_values(srv, cv)) {
		return HANDLER_ERROR;
	}
	
	if (p->conf_access_logfile->used) {
		if (-1 == (p->log_access_fd = 
			   open(p->conf_access_logfile->ptr, O_APPEND | O_WRONLY | O_CREAT | O_LARGEFILE, 0644))) {
			
			log_error_write(srv, __FILE__, __LINE__, "ss", 
					"opening access-log failed:", 
					strerror(errno));
			
			return HANDLER_ERROR;
		}
	} else {
		log_error_write(srv, __FILE__, __LINE__, "s", 
				"accesslog.filename has to be set");
		
		return HANDLER_ERROR;
	}
	
	return HANDLER_GO_ON;
}

SIGHUP_FUNC(log_access_cycle) {
	plugin_data *p = p_d;
	
	if (p->conf_access_logfile->used) {
		close(p->log_access_fd);
		
		if (-1 == (p->log_access_fd = 
			   open(p->conf_access_logfile->ptr, O_APPEND | O_WRONLY | O_CREAT | O_LARGEFILE, 0644))) {
			
			log_error_write(srv, __FILE__, __LINE__, "ss", "cycling access-log failed:", strerror(errno));
			
			return HANDLER_ERROR;
		}
	}
	
	return HANDLER_GO_ON;
}

REQUESTDONE_FUNC(log_access_write) {
	plugin_data *p = p_d;
	
#ifdef HAVE_IPV6
	int ndx, i;
#endif
	int newts = 0;
	data_string *ds;
	
	/* cache the generated timestamp */
	if (srv->cur_ts != p->last_generated_accesslog_ts) {
		struct tm tm;
#if defined(HAVE_STRUCT_TM_GMTOFF)
		long scd, hrs, min;
#endif
		
		buffer_prepare_copy(p->ts_accesslog_str, 255);
#if defined(HAVE_STRUCT_TM_GMTOFF)
		localtime_r(&(srv->cur_ts), &tm);
		strftime(p->ts_accesslog_str->ptr, p->ts_accesslog_str->size - 1, "%d/%b/%Y:%H:%M:%S ", &tm);
		p->ts_accesslog_str->used = strlen(p->ts_accesslog_str->ptr) + 1;
		
		buffer_append_string(p->ts_accesslog_str, tm.tm_gmtoff >= 0 ? "+" : "-");
		
		scd = abs(tm.tm_gmtoff);
		hrs = scd / 3600;
		min = (scd % 3600) / 60;
		
		/* hours */
		if (hrs < 10) buffer_append_string(p->ts_accesslog_str, "0");
		buffer_append_long(p->ts_accesslog_str, hrs);
		
		if (min < 10) buffer_append_string(p->ts_accesslog_str, "0");
		buffer_append_long(p->ts_accesslog_str, min);
#else
		gmtime_r(&(srv->cur_ts), &tm);
		strftime(p->ts_accesslog_str->ptr, p->ts_accesslog_str->size - 1, "%d/%b/%Y:%H:%M:%S +0000", &tm);
		p->ts_accesslog_str->used = strlen(p->ts_accesslog_str->ptr) + 1;
#endif
		
		p->last_generated_accesslog_ts = srv->cur_ts;
		newts = 1;
	}
	
	/* handle inet_ntop cache */
	
#ifdef HAVE_IPV6	
	ndx = 0;
	for (i = 0; i < INET_NTOP_CACHE_MAX; i++) {
		if (srv->inet_ntop_cache[i].ts != 0) {
			if (srv->inet_ntop_cache[i].family == AF_INET6 &&
			    0 == memcmp(srv->inet_ntop_cache[i].addr.ipv6.s6_addr, con->dst_addr.ipv6.sin6_addr.s6_addr, 16)) {
				/* IPv6 found in cache */
				break;
			} else if (srv->inet_ntop_cache[i].family == AF_INET &&
				   srv->inet_ntop_cache[i].addr.ipv4.s_addr == con->dst_addr.ipv4.sin_addr.s_addr) {
				/* IPv4 found in cache */
				break;
				
			}
		}
	}
	
	if (i == INET_NTOP_CACHE_MAX) {
		/* not found in cache */
		
		i = ndx;
		inet_ntop(con->dst_addr.plain.sa_family, 
			  con->dst_addr.plain.sa_family == AF_INET6 ? 
			  (const void *) &(con->dst_addr.ipv6.sin6_addr) :
			  (const void *) &(con->dst_addr.ipv4.sin_addr),
			  srv->inet_ntop_cache[i].b2, INET6_ADDRSTRLEN);
		
		srv->inet_ntop_cache[i].ts = srv->cur_ts;
		srv->inet_ntop_cache[i].family = con->dst_addr.plain.sa_family;
		
		if (srv->inet_ntop_cache[i].family == AF_INET) {
			srv->inet_ntop_cache[i].addr.ipv4.s_addr = con->dst_addr.ipv4.sin_addr.s_addr;
		} else if (srv->inet_ntop_cache[i].family == AF_INET6) {
			memcpy(srv->inet_ntop_cache[i].addr.ipv6.s6_addr, con->dst_addr.ipv6.sin6_addr.s6_addr, 16);
		}
	}
#endif
	
#ifdef HAVE_IPV6
	buffer_append_string(p->access_log, srv->inet_ntop_cache[i].b2);
#else
	buffer_append_string(p->access_log, inet_ntoa(con->dst_addr.ipv4.sin_addr));
#endif
	/* ident */
	BUFFER_APPEND_STRING_CONST(p->access_log, " - ");
	if (con->authed_user->used) {
		buffer_append_string_buffer(p->access_log, con->authed_user);
	} else {
		BUFFER_APPEND_STRING_CONST(p->access_log, "-");
	}
	BUFFER_APPEND_STRING_CONST(p->access_log, " [");
	
	buffer_append_string_buffer(p->access_log, p->ts_accesslog_str);
	BUFFER_APPEND_STRING_CONST(p->access_log, "] \"");
	
	if (con->request.request_line->used) {
		buffer_append_string_buffer(p->access_log, con->request.request_line);
	}
	BUFFER_APPEND_STRING_CONST(p->access_log, "\" ");
	
	buffer_append_long(p->access_log, con->http_status);
	
	BUFFER_APPEND_STRING_CONST(p->access_log, " ");
	
	if (con->bytes_written) {
		buffer_append_off_t(p->access_log, 
				    con->bytes_written - con->bytes_header <= 0 ? 0 : con->bytes_written - con->bytes_header);
	} else {
		BUFFER_APPEND_STRING_CONST(p->access_log, "-");
	}
	
	BUFFER_APPEND_STRING_CONST(p->access_log, " \"");
	
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Referer"))) {
		buffer_append_string_buffer(p->access_log, ds->value);
	}
	BUFFER_APPEND_STRING_CONST(p->access_log, "\" \"");
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "User-Agent"))) {
		buffer_append_string_buffer(p->access_log, ds->value);
	}
	
	/* duration */
#if 0
	BUFFER_APPEND_STRING_CONST(p->access_log, "\" ");
	buffer_append_long(p->access_log, srv->cur_ts - con->request_start);
	BUFFER_APPEND_STRING_CONST(p->access_log, "\n");
#else
	BUFFER_APPEND_STRING_CONST(p->access_log, "\"\n");
#endif
	
	if (newts || p->access_log->used > 256 * 1024) {
		write(p->log_access_fd, p->access_log->ptr, p->access_log->used - 1);
		buffer_reset(p->access_log);
	}
	
	return HANDLER_GO_ON;
}


int mod_accesslog_plugin_init(plugin *p) {
	p->name        = buffer_init_string("accesslog");
	
	p->init        = mod_accesslog_init;
	p->set_defaults= log_access_open;
	p->cleanup     = mod_accesslog_free;
	
	p->handle_request_done  = log_access_write;
	p->handle_sighup        = log_access_cycle;
	
	p->data        = NULL;
	
	return 0;
}
