#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "base.h"
#include "log.h"
#include "buffer.h"

#include "plugin.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

typedef struct {
	buffer *match_buf;
	
	pcre_keyvalue_buffer *conf_rewrite;
} plugin_data;

INIT_FUNC(mod_rewrite_init) {
	plugin_data *p;
	
	p = malloc(sizeof(plugin_data));
	memset(p, 0, sizeof(plugin_data));
	
	p->match_buf = buffer_init();
	
	p->conf_rewrite = pcre_keyvalue_buffer_init();
	
	return p;
}

FREE_FUNC(mod_rewrite_free) {
	plugin_data *p = p_d;
	
	UNUSED(srv);

	if (!p) return HANDLER_GO_ON;
	
	buffer_free(p->match_buf);
	pcre_keyvalue_buffer_free(p->conf_rewrite);
	
	free(p);
	
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC(mod_rewrite_set_defaults) {
	plugin_data *p = p_d;
	data_unset *du;
	size_t i = 0;
	
	config_values_t cv[] = { 
		{ "url.rewrite",                NULL, T_CONFIG_LOCAL }, /* 0 */
		{ NULL,                         NULL, T_CONFIG_UNSET }
	};
	
	if (!p) return HANDLER_ERROR;
	
	/* 0 */
	cv[i++].destination = p->conf_rewrite;
	
	if (0 != config_insert_values(srv, cv)) {
		return HANDLER_ERROR;
	}
	
	for (i = 0; cv[i].key; i++) {
		if (cv[i].type != T_CONFIG_LOCAL) continue;
		
		if (NULL != (du = array_get_element(srv->config, cv[i].key))) {
			if (du->type == TYPE_ARRAY) {
				size_t j;
				data_array *da = (data_array *)du;
				
				for (j = 0; j < da->value->used; j++) {
					if (da->value->data[j]->type == TYPE_STRING) {

						if (0 != pcre_keyvalue_buffer_append(cv[i].destination, 
									    ((data_string *)(da->value->data[j]))->key->ptr,
									    ((data_string *)(da->value->data[j]))->value->ptr)) {

							log_error_write(srv, __FILE__, __LINE__, "sb", "pcre-compile failed for", da->value->data[j]->key);
						}
					} else {
						log_error_write(srv, __FILE__, __LINE__, "sssbs", "unexpected type for key: ", cv[i].key, "[", da->value->data[i]->key, "](string)");
						
						return HANDLER_ERROR;
					}
				}
			} else {
				log_error_write(srv, __FILE__, __LINE__, "sss", "unexpected type for key: ", cv[i].key, "array of strings");
				
				return HANDLER_ERROR;
			}
		}
	}
	
	return HANDLER_GO_ON;
}


URIHANDLER_FUNC(mod_rewrite_uri_handler) {
#ifdef HAVE_PCRE_H
	plugin_data *p = p_d;
	size_t i;

	/* 
	 * REWRITE URL
	 * 
	 * e.g. rewrite /base/ to /index.php?section=base
	 * 
	 */
	
	buffer_copy_string_buffer(p->match_buf, con->request.uri);
	
	for (i = 0; i < p->conf_rewrite->used; i++) {
		pcre *match;
		const char *pattern;
		size_t pattern_len;
		int n;
# define N 10
		int ovec[N * 3];
		
		match = p->conf_rewrite->kv[i]->key;
		pattern = p->conf_rewrite->kv[i]->value;
		pattern_len = strlen(pattern);
		
		if ((n = pcre_exec(match, NULL, p->match_buf->ptr, p->match_buf->used - 1, 0, 0, ovec, 3 * N)) < 0) {
			if (n != PCRE_ERROR_NOMATCH) {
				log_error_write(srv, __FILE__, __LINE__, "sd",
						"execution error while matching: ", n);
				return HANDLER_ERROR;
			}
		} else {
			const char **list;
			size_t start, end;
			size_t k;
			
			/* it matched */
			pcre_get_substring_list(p->match_buf->ptr, ovec, n, &list);
			
			/* search for $[0-9] */
			
			buffer_reset(con->request.uri);
			
			start = 0; end = pattern_len;
			for (k = 0; k < pattern_len; k++) {
				if (pattern[k] == '$' &&
				    isdigit(pattern[k + 1])) {
					/* got one */
					
					size_t num = pattern[k + 1] - '0';
					
					end = k;
					
					buffer_append_string_len(con->request.uri, pattern + start, end - start);
					
					/* n is always larger than 0 */
					if (num < (size_t)n) {
						buffer_append_string(con->request.uri, list[num]);
					}
					
					k++;
					start = k + 1;
				} 
			}
			
			buffer_append_string_len(con->request.uri, pattern + start, pattern_len - start);
			
			pcre_free(list);
			
			return HANDLER_COMEBACK;
		}
	}
#undef N
		
#else
	UNUSED(srv);
	UNUSED(con);
	UNUSED(p_d);
#endif

	return HANDLER_GO_ON;
}

int mod_rewrite_plugin_init(plugin *p) {
	p->name        = buffer_init_string("rewrite");
	
	p->init        = mod_rewrite_init;
	p->handle_uri_raw = mod_rewrite_uri_handler;
	p->set_defaults = mod_rewrite_set_defaults;
	p->cleanup     = mod_rewrite_free;
	
	p->data        = NULL;
	
	return 0;
}
